package Yggdrasil;

use 5.008;
use strict;
use warnings;

### Revision 4.0.9 - 2003-09-08 - Morten Wulff <wulff@psyke.org>
$Yggdrasil::VERSION = '4.09';

$|++;

use AppConfig qw(:argcount);
use Carp;
use Date::Format;
use File::Basename;
use File::Copy;
use File::Find;
use File::Path;
use Getopt::Long;
use Net::FTP;
use Text::Template qw(fill_in_string);
use Text::Wrap;
use Yggdrasil::Cache;
use Yggdrasil::L10N;
use Yggdrasil::Progress::Bar;

###############################################################################
=pod

=head1 NAME

Yggdrasil - Generates a website from a directory tree of content and templates


=head1 SYNOPSIS


=head1 DESCRIPTION


=cut

###############################################################################
### INIT
###############################################################################

our $AUTOLOAD;

=pod

=head2 Configuration Variables

=over

=item input_dir

Path to source directory.

    input_dir => '/home/me/site/source'

=item output_dir

Path to output directory.

    output_dir => '/home/me/www'

=item data_dir

Directory used to store title and update time caches.

    data_dir => 'data'

=item dir_mode

Directory create mode used by mkpath().

    dir_mode => 0755

=item include

Files matching one or more of the regular expressions in this array are passed to Text::Template. Files that do not match any of the regular expressions are copied to the target directory without processing. Should only match HTML files.

    include => [qw(\.html$ \.shtml$)]

=item exclude

Files or directories matching one or more of the regular expressions in this array are not processed or copied. These regular expressions are applied before the regular expressions in the include array.

    exclude => [qw(^_ \.tmpl$ ~$)]

=item skip

Directories matching one or more of the regular expressions in this array are not included in the C<@navigation> array.

    skip => [qw(^img ^index$)]

=item base_url

Passed to templates.

    base_url => 'http://www.mysite.com'

=item base_template

Filename of base template. All parent directories of the current directory are searched.

    base_template => 'base.tmpl'

=item template_extension

File extension of template files.

    template_extension => '.tmpl'

=item template_override

A hashref containing full paths to override templates. If used there must be at least a 'default' override. Other overrides are matched against the current file name. E.g. an override template named 'beta' will be applied to all files matching /beta/.

    template_override => {beta => '/home/me/site/template.tmpl'}

=item timestamp_format

Format of the timestamp passed to the template.

    timestamp_format => '%a %b %e %Y (%T)',

=item html_indices

These filenames are removed from URL's in the breadcrumb trail. Should match your webserver configuration.

    html_indices => [qw(index.html index.shtml)]

=item html_only

Set to 1 to give all HTML files (the files in include) the extension .html (this is useful when creating site versions for offline use or for webservers that can't handle SSI).

    html_only => 0

=item front_page_navigation

Set to 1 to create a navigation array for the top level of the site (useful when creating site versions for offline use).

    front_page_navigation => 0

=item include_index

Set to 1 to include 'index.html' in navigation and breadcrumb trail links (useful when creating site versions for offline use).

    include_index => 0

=item siblings_include_index

Set to 1 to include index files in the siblings array.

    siblings_include_index => 0

=item home_label

Name given to the root page in the breadcrumb trail.

    home_label => 'Home'

=item title_tag

The tag in the source HTML document used as the document title.

    title_tag => 'h1'

=item default_title

Title given to files not found in the title cache.

    default_title => 'No Title'

=item subpages

Set to 1 to enable the numbered subpages feature (see User Guide).

    subpages => 1

=item pagination

Page title added to the breadcrumb trail for numbered subpages.

    pagination => 'Page '

=item ftp_host

URL or IP address of FTP server.

    ftp_host => 'ftp.mysite.com'

=item ftp_username

FTP server username.

    ftp_username => 'me'

=item ftp_password

FTP server password.

    ftp_password => 'secret'

=item ftp_path

Base path to use for remote files.

    ftp_path => '/www'

=item log_file_header

The first line(s) of the log file.

    log_file_header => "# Yggdrasil log file\n"

=item log_file_footer

The last line(s) of the log file.

    log_file_footer => '# end of log file'

=item log_file_format

Template for log file lines. The following variables are available: $fullname, $dirname, $filename.

    log_file_format => '{$fullname}'

=item release_file

Name of the file containing the last site release time. Stored in current data directory.

    release_file => '.release'

=item sp_command_line

Set this to the complete commandline for nsgmls if you wish to validate output files while processing. {$filename} is replaced with the current filename. You need to have the SP SGML system (http://www.jclark.com/sp/) installed in order to use this feature.

    sp_command_line => '/usr/bin/nsgmls {$filename}',

=item forcecache

Set to 1 to force rebuild of the document title cache.

    forcecache => 0

=item help

Set to 1 to display help message.

    help => 0

=item log_file

Name of log file.

    log_file => 'mysite.log'

=item update

Set to 1 to only process files added or changed since last release.

    update => 0

=item upload

Upload processed files via FTP.

    upload => 0

=item release

    release => 0

=item validate

    validate => 0

=item verbose

    verbose => 1

=back

=cut

### default settings - no trailing '/' on directory names and urls
my %Configuration = (
    ###########################################################################
    ### paths and such

    ### location of source files (content and templates)
    input_dir => '',

    ### location of finished site
    output_dir => '',

    ### store SDBM files here
    data_dir => 'data',

    ### directory create mode for mkpath
    dir_mode => 0755,

    ###########################################################################
    ### include/exclude filters

    ### include files matching these regexps, other files are copied
    ### (include is applied after exclude)
    include => [qw(\.html$ \.shtml$)],

    ### exclude files/paths matching these regexps
    exclude => [qw(^_ \.tmpl$ ~$)],

    ### exclude files/paths matching these regexps from navigation
    skip => [qw(^img ^index$)],

    ###########################################################################
    ### urls

    ### passed to templates and used for 'permanent location' links
    base_url => 'http://www.psyke.org',

    ###########################################################################
    ### templates

    ### base template
    base_template => 'base.tmpl',

    ### used when looking for templates in the current path
    template_extension => '.tmpl',

    ### override base template - see documentation for details
    template_override => {},

    ### format of timestamp passed to templates
    timestamp_format => '%a %b %e %Y (%T)',

    ###########################################################################
    ### navigation

    ### should match your web server configuration
    ### filenames in this array are stripped from generated URLs
    html_indices => [qw(index.html index.shtml)],

    ### set to 1 to give all HTML files the extension .html. this is useful
    ### when creating sites for offline use
    html_only => 0,

    ### set to 1 to create a navigation array for the top level of the site
    front_page_navigation => 0,

    ### set to 1 to include 'index.html' in navigation links. this is useful
    ### when creating sites for offline use
    include_index => 0,

    ### set to 1 to include index files in the siblings array
    siblings_include_index => 0,

    ### name of the root page (used in breadcrumb trails)
    home_label => 'Home',

    ### tag surrounding the document title
    ### title tag must be closed (i.e. <h1>title</h1> and on a single line)
    title_tag => 'h1',

    ### used for files where no title can be determined
    default_title => 'No Title',

    ### enable handling of subpages (see documentation for details)
    subpages => 1,

    ### used for content that spans multiple pages
    pagination => 'Page ',

    ###########################################################################
    ### ftp settings

    ftp_host     => '',
    ftp_username => '',
    ftp_password => '',
    ftp_path     => '',

    ###########################################################################
    ### log file

    ### log file header
    log_file_header => '# Yggdrasil log file',

    ### log file footer
    log_file_footer => '',

    ### format of log file line.
    ### available variables: $fullname, $filename, $dirname
    log_file_format => '{$fullname}',

    ###########################################################################

    ### file containing previous release time
    release_file => '.release',

    ### commandline used to invoke the nsgmls parser
    ### {$filename} is replaced with current output filename
    sp_command_line => '',

    ###########################################################################
    ### commandline switches and arguments override the following defaults

    config_file  => 'default', ### filename of configuration file
    forcecache   => 0,         ### set to 1 to force cache rebuild
    help         => 0,         ### print usage text
    log_file     => undef,     ### filename of log file
    update       => 0,         ### set to 1 to process only changed files
    upload       => 0,         ### set to 1 to upload changes files via FTP
    release      => 0,         ### set to 1 to reset release time
    validate     => 0,         ### validate changed HTML files
    verbose      => 1          ### verbosity level [0..2]
);

###############################################################################
### PUBLIC METHODS
###############################################################################

=pod

=head2 Public methods

=over

=cut

###############################################################################
=pod

=item go()

This method calls find() from File::Find. Depending on the configuration variables and commandline options it performs the following tasks:

Before calling C<find()>: Open log file for writing, flush output directory, initialize progress bar.

After calling C<find()>: Close log file, delete files from output directory, upload files via FTP, update release file.

=cut

sub go {
    my $self = shift;

    ### open log file
    if ($self->{log_file}) {
        open(LOG, ">$self->{log_file}")
            or die "Can't open '$self->{log_file}': $!\n";
        print LOG $self->{log_file_header}, "\n";
        $self->_pmt("Generating log file: '[_1]'\n", $self->{log_file})
            if $self->{verbose} > 0;
    }

    $self->_flush_output_directory() unless $self->{update};

    ### initialize progress bar
    if ($self->{verbose} > 0) {
        $self->_pmt("Processing '[_1]':\n\n", $self->{input_dir});
        $self->{progress_bar} =
            Yggdrasil::Progress::Bar->new($self->{cache}->size());
    }

    ### start processing directory tree
    find(
        {
            wanted      => $self->_bind_args(\&_process, $self),
            preprocess  => $self->_bind_args(\&_pre_process, $self),
            postprocess => $self->_bind_args(\&_post_process, $self)
        },
        $self->{input_dir}
    );
    print "\n\n";

    ### close log file
    if ($self->{log_file}) {
        print LOG $self->{log_file_footer};
        close(LOG);
    }

    $self->_delete_files()        if $self->{update};
    $self->_upload_files()        if $self->{upload};
    $self->_update_release_file() if $self->{release};
}

###############################################################################
=pod

=item new()

Creates a new Yggdrasil object and initializes it with values from the %Configuration hash and the configuration file if applicable.

See also: L<C<_init()>|/_init>

=cut

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my %parameters = @_;

    ### initialize object
    my $self  = {
        _permitted => \%Configuration
    };
    foreach my $element (keys %Configuration) {
        $self->{$element} = $parameters{$element} || $Configuration{$element};
    }

    bless ($self, $class);

    $self->_init(); ### complete object initialization

    return $self;
}

###############################################################################
=pod

=item stats()

Returns a string with site processing stats.

=cut

sub stats {
    my $self = shift;

    ### convert sizes to kilobytes
    $self->{stats}->{total_size} = $self->{stats}->{total_size} > 1024
        ? int($self->{stats}->{total_size}/1024) . " KiB"
        : $self->{stats}->{total_size} . " B";
    $self->{stats}->{html_size}  = $self->{stats}->{html_size} > 1024
        ? int($self->{stats}->{html_size}/1024) . "KiB"
        : $self->{stats}->{html_size} . " B";

    ### calculate run time
    $self->{stats}->{elapsed_time} = $self->{stats}->{start_time} + time;

    ### build processing stats
    my $stats = $self->{l10n}->maketext(
        '_stats',
        map {$self->{stats}->{$_}} @{$self->{l10n}->maketext('_stats_order')}
    );

    ### append changes and errors if they exist
    my $changes = {
        added   => [],
        changed => [],
        deleted => []
    };
    foreach my $file (keys %{$self->{processed}}) {
        push @{$changes->{$self->{processed}->{$file}}}, $file;
    }

    $stats .= $self->_rmt(
        '_stats_added',
        join "\n", @{$changes->{added}}
    ) if @{$changes->{added}};
    $stats .= $self->_rmt(
        '_stats_changed',
        join "\n", @{$changes->{changed}}
    ) if @{$changes->{changed}};
    $stats .= $self->_rmt(
        '_stats_deleted',
        join "\n", @{$changes->{deleted}}
    ) if @{$changes->{deleted}};

    $stats .= $self->_rmt(
        '_stats_errors',
        $self->{stats}->{errors}
    ) if $self->{stats}->{errors};

    return $stats;
}

=pod

=back

=cut

###############################################################################
### PRIVATE METHODS
###############################################################################

=pod

=head2 Private methods

=over

=cut

###############################################################################
=pod

=item _pre_process(@files)

Determines which files get sent to the L<C<_process()>|/_process> method. Creates or updates the following arrays:

I<@siblings>: A list of all files in the current directory, except subdirectories and HTML index files (set the configuration variable C<siblings_include_index> to 1 to include indexes).

I<@trail>: The path to the current directory. The title of the root page is determined by the C<home_label> configuration variable. If the name of the current directory matches the name of the parent directory and the name of the current directory matches C<.*\/(\d+)$>, the current page is treated as a numbered subpage and a new title is generated from the C<pagination> configuration variable.

I<@navigation>: A list of subpages. Contains all subpages except those matching any of the regexps in C<skip> or those that are numbered subpages of the current page.

All the above arrays contain hashrefs with the following data:

    {
        name => 'Page Title',
        href => 'relative_path/'
    }

See also: L<C<_process()>|/_process>

=cut

sub _pre_process {
    my $self  = shift;
    my @files = sort { $a cmp $b } @_;

    my $dir = $File::Find::dir;

    ### remove unwanted elements from the file list
    @files = grep {
        my $file = $_;
        my $path = join '/', $dir, $file;
        !/^\.{1,2}/ and
        !grep {$file =~ /$_/ or $path =~ /$_/} @{$self->{exclude}}
    } @files;

    ### create list of files in the current directory
    @{$self->{siblings}} = map {
        {
            name => $self->{cache}->title(join '/', $dir, $_) || $_,
            href => $_
        }
    } grep {
        my $file = $_;
        !-d and
        !(!$self->{siblings_include_index} and
          grep {$file =~ /^$_$/} @{$self->{html_indices}})
    } @files;

    ### update breadcrumb trail
    my $name = $self->{cache}->title($dir);
    if ($#{$self->{trail}} > -1) {
        ### handle numbered subpages (/2 /3 etc.)
        if ($self->{subpages} and
            $name eq $self->{trail}->[-1]->{name} and
            $dir =~ /.*\/(\d+)$/) {
            $name = join '', $self->{pagination}, $1;
        }
        push @{$self->{trail}}, {name => $name};
        $self->{stats}->{directories}++;
    }
    else {
        ### root page
        push @{$self->{trail}}, {name => $name};
    }

    ### create links to subdirectories (except those in $self->{skip} or those
    ### with subpages)
    if ($#{$self->{trail}} > 0 || $self->{front_page_navigation}) {
        @{$self->{navigation}} = map {
            {
                name => $self->{cache}->title(join '/', $dir, $_),
                href => $self->{include_index} ? "$_/index.html" : "$_/"
            }
        } grep {
            my $file = $_;
            my $path = join '/', $dir, $file;

            ### include directories
            -d and

            ### exclude subpages with the same name as their parent and a name
            ### with only digits unless the subpages option is set
            !($self->{subpages} and
              $self->{cache}->title($path) eq $self->{trail}->[-1]->{name} and
              $file =~ /\d+/) and

            ### exclude files/paths matching any of the regexps in skip
            !grep {$file =~ /$_/ || $path =~ /$_/} @{$self->{skip}}
        } @files;
    }

    $self->{depth}++;

    return @files;
}

###############################################################################
=pod

=item _process()

Description of _process().

=cut

sub _process {
    my $self = shift;

    return if -d;

    $self->{progress_bar}->update() if $self->{verbose} > 0;

    ### handle added/updated files and use of the --update switch
    my $touch = 0;
    $self->{processed}->{$File::Find::name} = 'none';
    if ($self->{cache}->has_been_added($File::Find::name)) {
        ### file has been added to the cache
        $self->{processed}->{$File::Find::name} = 'added';
        $touch++;
    }
    else {
        ### file is already in the cache
        if ((stat($File::Find::name))[9] > $self->{release_time}) {
            ### file has been modified
            $self->{processed}->{$File::Find::name} = 'changed';
        }
        else {
            ### file has not been modified
            return if $self->{update};
        }
    }

    ### set target file
    my $target = $File::Find::name;
    $target =~ s/$self->{input_dir}/$self->{output_dir}/o;

    $self->_create_directory($target);

    ### update stats
    my $file_size = (stat $File::Find::name)[7];
    $self->{stats}->{total_size} += $file_size;
    $self->{stats}->{files}++;

    ### choose process or copy mode
    if ($self->_include($File::Find::name)) {
        $self->{progress_bar}->marker('P>') if $self->{verbose} > 0;

        $self->_print_logline($File::Find::name) if $self->{log_file};

        ### modify trail if this isn't an index file
        my $non_index = 0;
        if (not $self->_html_index($File::Find::name)) {
            push @{$self->{trail}}, {
                name => $self->{cache}->title($File::Find::name)
            };
            $non_index = 1;
        }

        ### get templates in current path (makes including templates easier)
        $self->_templates_in_path($File::Find::dir);

        ### template override in effect
        if (%{$self->{template_override}}) {
            $self->{template}->{__base__} = '';
            foreach my $pattern (keys %{$self->{template_override}}) {
                next if $pattern eq 'default';
                if ($File::Find::name =~ m/$pattern/) {
                    $self->{template}->{__base__} =
                        $self->{template_override}->{$pattern};
                }
            }
            unless ($self->{template}->{__base__}) {
                $self->{template}->{__base__} =
                    $self->{template_override}->{default};
            }
        }

        die "No base template found!\n" unless $self->{template}->{__base__};

        ### template stuff
        my $template = Text::Template->new(
            TYPE       => 'FILE',
            SOURCE     => $self->{template}->{__base__}
        ) or die "$Text::Template::ERROR";

        $target =~ s/(.*)\.(.*)/$1\.html/ if $self->{html_only};

        $self->{template}->{__file__} = $File::Find::name;
        $self->_update_content($File::Find::name);

        open(OUT, ">$target") or die "Can't open '$target': $!\n";
        if ($self->{external_post_process}) {
            my $output = $template->fill_in(
                PACKAGE => 'Yggdrasil::Content'
            ) or warn "$Text::Template::ERROR";
            $output =
                &{$self->{external_post_process}}($output, $self->{exclude});
            print OUT $output;
        }
        else {
            $template->fill_in(
                PACKAGE => 'Yggdrasil::Content',
                OUTPUT => \*OUT
            ) or warn "$Text::Template::ERROR";
        }
        close(OUT);

        $self->_validate_file($target) if $self->{validate};

        pop @{$self->{trail}} if $non_index;

        $self->{stats}->{html_size} += $file_size;
    }
    else {
        ### make sure that the file shows as updated in subsequent runs
        if ($touch) {
            $self->{cache}->update($File::Find::name);
            utime time, time, $File::Find::name;
        }

        $self->{progress_bar}->marker('C>') if $self->{verbose} > 0;

        copy($File::Find::name, $target)
           or die "Can't copy '$File::Find::name' to '$target': $!\n";
    }
}

###############################################################################
=pod

=item _post_process()

Removes the current directory from the breadcrumb trail and decrements C<$depth>.

=cut

sub _post_process {
    my $self = shift;

    pop @{$self->{trail}};
    $self->{depth}--;
}



###############################################################################
=pod

=item _bind_args()

Makes it possible to pass an object reference to the File::Find functions.

See also: http://perlmonks.thepen.com/109068.html

=cut

sub _bind_args {
    my $self = shift;
    my ($sub, @args) = @_;

    return sub {$sub->(@args, @_);};
}

###############################################################################
=pod

=item _create_directory($directory)

Creates the directory if it doesn't exist. Uses the directory create mode defined in C<dir_mode>.

=cut

sub _create_directory {
    my $self = shift;
    my $dir  = dirname(shift);

    if (not -d $dir) {
        mkpath($dir , 0, $self->{dir_mode})
            or die "Can't create directory '$dir': $!\n";
    }
}

###############################################################################
=pod

=item _delete_files()

Calls $cache->sync() to get a list of the files removed from the input directory (if any). Deletes those files from the output directory.

=cut

sub _delete_files {
    my $self = shift;

    my @deleted = $self->{cache}->sync();

    foreach my $file (@deleted) {
        $self->{processed}->{$file} = 'deleted';
        my $delete_file = $file;
        $delete_file =~ s/$self->{input_dir}/$self->{output_dir}/;
        unless (unlink $delete_file) {
            warn "Can't delete file '$delete_file': $!";
        }
    }
}

###############################################################################
=pod

=item _flush_output_directory()

Removes all content from the output directory.

=cut

sub _flush_output_directory {
    my $self = shift;

    $self->_pmt("Flushing output dir...") if $self->{verbose} > 0;

    if (-d $self->{output_dir}) {
        rmtree($self->{output_dir})
            or die "Can't remove directory '$self->{output_dir}': $!\n";
    }
    mkpath($self->{output_dir}, 0, $self->{dir_mode})
        or die "Can't create directory '$self->{output_dir}': $!\n";

    $self->_pmt(" Done!\n") if $self->{verbose} > 0;
}

###############################################################################
=pod

=item _get_release()

Returns last release time from C<release_file>. Sets C<release_time> to 0 if release file doesn't exist.

=cut

sub _get_release {
    my $self = shift;

    ### get time of last release or set default value
    if (-e $self->{release_file}) {
        open(IN, $self->{release_file})
            or die "Can't open '$self->{release_file}': $!\n";
        my $release = <IN>;
        close(IN);

        if ($release =~ /^\d+$/) {
            $self->{release_time} = $release;
            return;
        }
        else {
            $self->_pmt(
                "Release file '[_1]' corrupted.\n",
                $self->{release_file}
            ) if $self->{verbose} > 0;
        }
    }
    else {
        $self->_pmt(
            "Release file '[_1]' doesn't exist.\n",
            $self->{release_file}
        ) if $self->{verbose} > 0;
    }

    ### set default release time and force creation of new release file
    $self->{release_time} = 0;
    $self->{release} = 1;
}

###############################################################################
=pod

=item _html_index($filename)

Returns true if the file is an HTML index file, false otherwise.

=cut

sub _html_index {
    my $self = shift;
    my $file = basename(shift);

    return scalar(grep {$file =~ /^$_$/} @{$self->{html_indices}});
}

###############################################################################
=pod

=item _include($filename)

Returns true if C<$filename> matches any of the regexps in C<include>, false otherwise.

=cut

sub _include {
    my $self = shift;
    my $file = shift;

    return scalar(grep {$file =~ /$_/} @{$self->{include}});
}

###############################################################################
=pod

=item _init()

Description of init.

=cut

sub _init {
    my $self = shift;

    ### initialize localization
    $self->{l10n} = Yggdrasil::L10N->get_handle()
        or die "Can't find language file";

    ### get command-line options
    GetOptions(
        'config|c=s'   => \$self->{config_file},
        'cache'        => \$self->{forcecache},
        'help|?'       => \$self->{help},
        'log|l:s'      => \$self->{log_file},
        'update|u'     => \$self->{update},
        'upload|ftp|f' => \$self->{upload},
        'release|r'    => \$self->{release},
        'validate|v'   => \$self->{validate},
        'verbose=i'    => \$self->{verbose}
    ) or do {$self->_pmt('_usage'); exit(2)};

    $self->_read_configuration_file();

    ### print help message
    do {$self->_pmt('_help'); exit(1)} if $self->{help};

    ### set default log file if none is given
    $self->{log_file} ||= 'yggdrasil.log' if defined($self->{log_file});

    $self->_get_release();
    $self->_validate_configuration();

    ### initialize stat counters and arrays
    $self->_init_stats();

    $self->{processed} = {};

    ### initialize navigation
    $self->{trail}      = [];
    $self->{navigation} = [];

    $self->{depth} = -1;

    ### initialize cache
    $self->{cache} = Yggdrasil::Cache->new(
        data_dir  => $self->{data_dir},
        input_dir => $self->{input_dir},
        include   => $self->{include},
        exclude   => $self->{exclude},
        title_tag => $self->{title_tag},
        force     => $self->{forcecache},
        verbose   => $self->{verbose}
    );

    ### initialize template utility functions
    *Yggdrasil::Content::include = sub {
        my $file = shift;
        Text::Template::fill_in_file($file, PACKAGE => 'Yggdrasil::Content');
    };
    *Yggdrasil::Content::include_text = \&Text::Template::_load_text;
}

###############################################################################
=pod

=item _init_stats()

Resets the counters used to gather processing statistics.

=cut

sub _init_stats {
    my $self = shift;

    $self->{stats} = {
        files        => 0,
        directories  => 0,
        total_size   => 0,
        html_size    => 0,
        start_time   => -time,
        elapsed_time => 0
    };
}

###############################################################################
=pod

=item _pmt($lexicon_entry,@parameters)

Wrapper for the maketext function, prints the result from C<maketext()> (Print MakeText).

=cut

sub _pmt {
    my $self = shift;

    print $self->{l10n}->maketext(@_);
}

###############################################################################
=pod

=item _print_logline()

Prints formatted line to logfile.

=cut

sub _print_logline {
    my $self = shift;
    my $file = shift;

    my $log_line = fill_in_string(
        $self->{log_file_format},
        HASH => {
            dirname  => dirname($file),
            filename => basename($file),
            fullname => $file
        }
    );
    print LOG "$log_line\n";
}

###############################################################################
=pod

=item _read_configuration_file()

Reads the configuration file specified on the commandline or the default configuration file. Settings in the configuration file override the defaults in C<%Configuration>.

Creates cache directory if it doesn't exist.

Initializes external postprocessing sub if present.

=cut

sub _read_configuration_file {
    my $self = shift;

    ### check if configuration file exists
    my $config_file = join '', 'conf', '/', $self->{config_file}, '.cfg';
    unless (-e $config_file) {
        die "Configuration file '$config_file' doesn't exist.\n";
    }

    ### initialize AppConfig
    my $config = AppConfig->new({
        CASE     => 1,
        PEDANTIC => 1
    });
    foreach my $key (keys %Configuration) {
        ref($Configuration{$key}) eq 'ARRAY' && do {
            $config->define( $key => {ARGCOUNT => ARGCOUNT_LIST} );
            next;
        };
        ref($Configuration{$key}) eq 'HASH' && do {
            $config->define( $key => {ARGCOUNT => ARGCOUNT_HASH} );
            next;
        };
        $config->define( $key => {ARGCOUNT => ARGCOUNT_ONE} );
    }

    ### read configuration file
    $config->file($config_file);

    ### settings from the configuration file override the defaults in $self
    my %varlist = $config->varlist('.*');
    foreach my $key (keys %varlist)
    {
        if (ref($config->get($key)) eq 'ARRAY') {
            $self->{$key} = $config->get($key) if @{$config->get($key)};
        }
        else {
            $self->{$key} = $config->get($key) if defined($config->get($key));
        }
    }

    ### create separate cache directories for alternative configuratons
    if ($self->{config_file} ne 'default') {
        $self->{data_dir} = join '/', $self->{data_dir}, $self->{config_file};
    }
    $self->{release_file} = join '/', $self->{data_dir}, $self->{release_file};

    ### initialize external postprocessing if present
    my $sub_file = join '', 'conf', '/', $self->{config_file}, '.sub';
    if (-e $sub_file) {
        $self->{external_post_process} = do($sub_file)
            or die "Can't initialize post processing '$sub_file': $!\n";
    }
}

###############################################################################
=pod

=item _rmt($lexicon_entry,@parameters)

Wrapper for the maketext function, returns the result from C<maketext()> (Return MakeText).

=cut

sub _rmt {
    my $self = shift;

    return $self->{l10n}->maketext(@_);
}

###############################################################################
=pod

=item _templates_in_path($directory)

Searches the current path for files with the extension C<template_extension>. Templates in subdirectories override templates in the root directory.

The special template C<__base__> is set according to the configuration variable C<base_template>.

=cut

sub _templates_in_path {
    my $self = shift;
    my $dir  = shift;

    my @path = split /\//, $dir;

    $self->{template} = {};

    while (@path) {
        my $dir = join '/', @path;

        opendir(DIR, $dir) or die "Can't open '$dir': $!";
        while(my $entry = readdir(DIR)) {
            next unless $entry =~ /$self->{template_extension}$/;

            my $template_name = basename($entry);
            my $template_path = join '/', @path, $template_name;

            $self->{template}->{$template_name} ||= $template_path;
        }
        closedir(DIR);

        last if $dir eq $self->{input_dir};

        pop @path;
    }

    $self->{template}->{__base__} =
        $self->{template}->{$self->{base_template}};
}

###############################################################################
=pod

=item _trail_links()

Creates relative links to the pages in the breadcrumb trail. Fixes links if the current file is not an HTML index file.

See also: L<C<_html_index()>|/_html_index>

=cut

sub _trail_links {
    my $self = shift;

    my $dots = $#{$self->{trail}};

    ### insert home label or use root file title
    $self->{trail}->[0]->{name} = $self->{home_label} if $self->{home_label};

    ### create relative links to pages in breadcrumb trail
    foreach my $item (@{$self->{trail}}) {
        $item->{href}  = $dots ? '../' x $dots : '';
        $item->{href} .= 'index.html' if ($self->{include_index} && $dots);
        $dots--;
    }

    ### fix trails for non-index html files
    unless ($self->_html_index($File::Find::name)) {
        my $index_pos = $#{$self->{trail}}-1;
        my $count = 0;
        foreach (@{$self->{trail}}) {
            if ($count == $index_pos) {
                $_->{href} = './';
            }
            else {
                $_->{href} =~ s!../!!;
            }
            $count++;
        }
    }
}

###############################################################################
=pod

=item _update_content($filename)

Prepares the content used by the templating system.

=cut

sub _update_content {
    my $self = shift;
    my $file = shift;
    my $dir  = dirname($file);

    $self->_trail_links();

    my $location = $dir;
    $location =~ s/$self->{input_dir}/$self->{base_url}/o;

    ### make sure that permanent location works for non-index files
    my $filename = basename($file);
    unless ($self->_html_index($filename)) {
        $location .= "/$filename" ;
    }

    @Yggdrasil::Content::navigation = @{$self->{navigation}};
    @Yggdrasil::Content::siblings   = @{$self->{siblings}};
    @Yggdrasil::Content::trail      = @{$self->{trail}};

    $Yggdrasil::Content::base_url  = join '', $self->{base_url}, '/';
    $Yggdrasil::Content::depth     = $self->{depth};
    $Yggdrasil::Content::location  = $location;
    $Yggdrasil::Content::rel_path  = '../' x ( $self->{depth} );
    $Yggdrasil::Content::timestamp = time2str($self->{timestamp_format}, time);
    $Yggdrasil::Content::title     = $self->{cache}->title($file);

    %Yggdrasil::Content::template = %{$self->{template}};
}

###############################################################################
=pod

=item _update_release_file()

Sets the time in the release file to the current time.

=cut

sub _update_release_file {
    my $self = shift;

    if ($self->{release}) {
        $self->_pmt("Updating release file...") if $self->{verbose} > 0;
        open(RELEASE, ">$self->{release_file}")
            or die "Can't open '$self->{release_file}': $!";
        print RELEASE time;
        close(RELEASE);
        $self->_pmt(" Done!\n\n") if $self->{verbose} > 0;
    }
}

###############################################################################
=pod

=item _upload_files()

Upload processed files to the FTP server defined in the configuration file.

=cut

sub _upload_files {
    my $self = shift;

    ### open ftp connection
    my $ftp = Net::FTP->new($self->{ftp_host}, Debug => 0) or die $@;
    $ftp->login($self->{ftp_username}, $self->{ftp_password})
        or die "Login failed";

    ### create list of files to upload
    my @upload;
    if ($self->{update}) {
        ### only upload added or changed files
        @upload = grep {
            $self->{processed}->{$_} eq 'added' or
            $self->{processed}->{$_} eq 'changed'
        } keys %{$self->{processed}};
    }
    else {
        ### upload all files
        @upload = grep {
            $self->{processed}->{$_} ne 'deleted'
        } keys %{$self->{processed}};
    }

    ### upload files
    foreach my $file (@upload) {
        my $local_file  = $file;
        my $remote_file = $file;

        $local_file  =~ s/$self->{input_dir}/$self->{output_dir}/;
        $remote_file =~ s/$self->{input_dir}//;
        $remote_file = $self->{ftp_path} . $remote_file;

        ### create remote directory if it doesn't exist
        my $remote_path = dirname($remote_file);
        unless ($ftp->ls($remote_path)) {
            $ftp->mkdir($remote_path, 1);
        }

        $self->_pmt("FTP PUT [_1]...", $remote_file) if $self->{verbose} > 1;

        ### set data transfer type
        if ($self->_include($file)) {
            $ftp->ascii();
        }
        else {
            $ftp->binary();
        }

        $ftp->put($local_file, $remote_file)
            or warn "Can't put '$remote_file': $!";

        $self->_pmt(" Done!\n", $remote_file) if $self->{verbose} > 1;
    }
    print "\n" if $self->{verbose} > 1;

    ### delete files that are no longer in the source tree
    ### FIXME - delete dirs
    my @deleted = grep {
        $self->{processed}->{$_} eq 'deleted'
    } keys %{$self->{processed}};
    foreach my $file (@deleted) {
        my $remote_file = $file;
        $remote_file =~ s/$self->{input_dir}//;
        $remote_file = $self->{ftp_path} . $remote_file;

        $ftp->delete($remote_file) or warn "Can't delete '$remote_file'"

        ### if dirname(cwd) is empty -> rmdir(cwd) (we are still in the root)
    }

    ### close ftp connection
    $ftp->quit;
}

###############################################################################
=pod

=item _validate_configuration()

Check configuration files for basic errors.

=cut

sub _validate_configuration {
    my $self = shift;

    warn "No file inclusion regular expressions specified"
        unless $#{$self->{include}} >= 0;
    warn "No file exclusion regular expressions specified"
        unless $#{$self->{exclude}} >= 0;
    warn "Input directory '$self->{input_dir}' doesn't exist"
        unless -d $self->{input_dir};
    warn "No base URL specified"      unless $self->{base_url};
    warn "No base template specified" unless $self->{base_template};
    warn "No HTML indices specified"  unless $#{$self->{html_indices}} >= 0;

    ### check validator
    if ($self->{validate}) {
        my $executable = (split /\s+/, $self->{sp_command_line})[0];
        die "Validator '$executable' doesn't exist." unless -x $executable;
    }
}

###############################################################################
=pod

=item _validate_file($filename)

Inserts C<$filename> in C<sp_command_line> and executes the command.

Any error messages returned by the validator are appended to the C<errors> string.

=cut

sub _validate_file  {
    my $self = shift;
    my $file = shift;

    ### insert filename in sp_command_line
    my $validate = fill_in_string(
        $self->{sp_command_line},
        HASH => {filename => $file}
    );

    ### execute command
    open(SP, "$validate 2>&1 |") or die "Can't run validator: $!";
    my @result = <SP>;

    foreach (@result) {
        next unless /$file/;

        my ($line, $char, $type, $mesg) = /$file:(\d+):(\d+):(\w+)*:*\s+(.*)/;

        $type ||= 'M';
        $file =~ s/$self->{output_dir}//;

        $self->{stats}->{errors} .= "$file ($line, $char):\n";
        $self->{stats}->{errors} .= wrap("\t", "\t", "$mesg\n");
    }
}

=pod

=back

=cut

###############################################################################
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) or die "$self is not an object";

    my $name = $AUTOLOAD;
    $name =~ s/.*://;

    unless (exists $self->{_permitted}->{$name} ) {
        croak "Can't access '$name' field in class $type";
    }

    if (@_) {
        return $self->{$name} = shift;
    }
    else {
        return $self->{$name};
    }
}

1;

__END__

=pod

=head1 BUGS/CAVEATS/etc

No known bugs.


=head1 REQUIREMENTS

Yggdrasil requires the following modules:

C<AppConfig>

C<Date::Format>

C<Text::Template>


=head1 AUTHOR

Morten Wulff, <wulff@psyke.org>


=head1 SEE ALSO

C<Text::Template>


=head1 COPYRIGHT

Copyright 2002-2003, Morten Wulff

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut
