###############################################################################
### Cache.pm (Yggdrasil::Cache)
### Manages titles and update times for a directory of HTML files
###
### Copyright (C) 2003 Morten Wulff <wulff@psyke.org>
###
### This program is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License
### as published by the Free Software Foundation; either version 2
### of the License, or (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this program; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###############################################################################

### REVISIONS
### 1.0.0 : 2003-04-11 : Morten Wulff <wulff@psyke.org>
### 1.1.0 : 2003-04-24 : Morten Wulff <wulff@psyke.org>
###         Improved interface, easier to include/exclude files
### 1.1.1 : 2003-04-26 : Morten Wulff <wulff@psyke.org>
###         Minor code clean-up
### 1.1.2 : 2003-04-27 : Morten Wulff <wulff@psyke.org>
###         Fixed problem in _unlink_cache
### 1.2.0 : 2003-05-22 : Morten Wulff <wulff@psyke.org>
###         Fixed title() cache update behavior
### 1.3.0 : 2003-08-21 : Morten Wulff <wulff@psyke.org>
###         Changed verbosity behavior
###         Fixed dir create mode (Thomas Eskebk)
### 1.4.0 : 2003-08-26 : Morten Wulff <wulff@psyke.org>
###         Added cached(), delete() and keys_update() methods
### 1.4.1 : 2003-08-28 : Morten Wulff <wulff@psyke.org>
###         Added has_been_added(), removed cached()
### 1.4.2 : 2003-08-30 : Morten Wulff <wulff@psyke.org>
###         Added config option title_tag, made tag matching case-insensitive

package Yggdrasil::Cache;

$Yggdrasil::Cache::VERSION = '1.42';

use strict;
use warnings;

$|++;

use Carp;
use Fcntl;
use File::Basename;
use File::Find;
use File::Path;
use SDBM_File;

###############################################################################
### INIT
###############################################################################

our $AUTOLOAD;

my %Fields = (
    ###########################################################################
    ### paths and such

    ### where to start crawling
    input_dir => '',

    ### store SDBM files here
    data_dir  => '',

    ### dir create mode for mkpath
    dir_mode  => 0755,

    ###########################################################################
    ### include/exclude filters

    ### include files matching these regexp's (include is applied before exclude)
    include => [qw(\.html$ \.shtml$)],

    ### exclude files/paths matching these regexp's
    exclude => [qw(^_ \.tmpl$)],

    ###########################################################################
    ### navigation

    ### should match your apache config (are stripped from generated URL's)
    html_indices => [qw(index.html index.shtml)],

    ### tag surrounding the document title
    title_tag => 'h1',

    ### return for HTML files not in cache
    default_title => 'No Title',

    ###########################################################################

    ### force rebuild of cache (old database files are deleted)
    force   => 0,

    ### verbosity level [0..2]
    verbose => 1
);

###############################################################################
### PUBLIC METHODS
###############################################################################

###############################################################################
sub cached {
    my $self     = shift;
    my $filename = shift;

    return exists $self->{update}->{$filename};
}

###############################################################################
sub delete {
    my $self     = shift;
    my $filename = shift;

    delete $self->{update}->{$filename};
}

###############################################################################
sub has_been_added {
    my $self     = shift;
    my $filename = shift;

	return scalar(grep {$_ eq $filename} @{$self->{added}});
}

###############################################################################
sub keys_update {
    my $self = shift;

    return keys %{$self->{update}};
}

###############################################################################
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my %parameters = @_;

    ### initialize object
    my $self  = {
        _permitted => \%Fields
    };
    foreach my $element (keys %Fields) {
        $self->{$element} = $parameters{$element} || $Fields{$element};
    }

    ### check for presence of input directory
    unless ( -d $self->{input_dir} ) {
        croak "Input directory '$self->{input_dir}' doesn't exist";
    }

    bless ($self, $class);

    ### tie cache files and perform updates
    $self->_init();

    return $self;
}

###############################################################################
sub size {
    my $self = shift;

    return scalar(keys %{$self->{update}});
}

###############################################################################
sub sync {
    my $self = shift;

	my @deleted;
	foreach my $filename (keys %{$self->{update}}) {
		unless (-e $filename) {
			push @deleted, $filename;
		    delete $self->{title}->{$filename};
		    delete $self->{update}->{$filename};
		}
	}

	return @deleted
}

###############################################################################
sub title {
    my $self     = shift;
    my $filename = shift;

    unless (exists $self->{title}->{$filename}) {
        if (-e $filename) {
            SWITCH: {
                -d $filename and do { ### $filename is a directory
                    my $directory = $filename;

                    ### grab the first existing index file
                    my ($file) = grep {-f join '/', $directory, $_} @{$self->{html_indices}};
                    if ($file) {
                        $self->_process_file($file, $directory);
                    }
					push @{$self->{added}}, $filename if $file;
                    last SWITCH;
                };
                -f $filename and do { ### $filename is a plain file
                    my $directory = dirname($filename);
                    my $file      = basename($filename);

					$self->_process_file($file, $directory);
					push @{$self->{added}}, $filename if $file;
                    last SWITCH;
                };

                ### fallthrough
                carp "Cache: '$filename' is not a plain file or a directory";
            }
        }
        else {
            carp "Cache: '$filename' doesn't exist";
            return;
        }
    }

    return exists $self->{title}->{$filename}
        ? $self->{title}->{$filename}
        : $self->{default_title};
}

###############################################################################
sub update {
    my $self     = shift;
    my $filename = shift;

    unless (exists $self->{update}->{$filename}) {
        if (-e $filename) {
            my $directory = dirname($filename);
            $self->_process_file($filename, $directory);
        }
        else {
            carp "Cache: '$filename' doesn't exist";
            return;
        }
    }

    return exists $self->{update}->{$filename} ? $self->{update}->{$filename} : 0;
}

###############################################################################
### PRIVATE METHODS
###############################################################################

###############################################################################
### http://perlmonks.thepen.com/109068.html
sub _bind_args {
    my $self = shift;
    my ($sub, @args) = @_;

    return sub {$sub->(@args, @_);};
}

###############################################################################
sub _create_dir {
    my $self = shift;

    ### force cache rebuild
    $self->{force} = 1;

    print "Creating cache directory '$self->{data_dir}'\n" if $self->{verbose} > 0;

    mkpath($self->{data_dir}, 0, $self->{dir_mode}) or croak "Cache: Cannot create cache directory '$self->{data_dir}': $!";
}

###############################################################################
sub _init {
    my $self = shift;

    ### create cache directory if it doesn't exist
    $self->_create_dir() unless -d $self->{data_dir};

    ### force cache rebuild if data files are missing
    unless (
        -e "$self->{data_dir}/title.dir"  and
        -e "$self->{data_dir}/title.pag"  and
        -e "$self->{data_dir}/update.dir" and
        -e "$self->{data_dir}/update.pag"
    ) {
        print "Forcing cache rebuild\n" if $self->{verbose} > 0;
        $self->{force} = 1;
    }

    $self->_unlink_cache() if $self->{force};
    $self->_tie_cache();

    if ($self->{force}) {
        print "Building cache in '$self->{data_dir}'... " if $self->{verbose} > 0;

        find(
            {
                wanted      => $self->_bind_args(\&_process, $self),
                preprocess  => $self->_bind_args(\&_pre_process, $self)
            },
            $self->{input_dir}
        );

        print "Done!\n" if $self->{verbose} > 0;
    };

	$self->{added} = [];
}

###############################################################################
sub _pre_process {
    my $self  = shift;
    my @files = @_;

    ### remove unwanted elements from the file list
    @files = grep {
        my $file = $_;
        my $path = join '/', $File::Find::dir, $file;
        !/^\.{1,2}/ and
        !grep {$file =~ /$_/ || $path =~ /$_/} @{$self->{exclude}}
    } @files;

    return @files;
}

###############################################################################
sub _process {
    my $self = shift;

    return if -d;

    $self->_process_file($_, $File::Find::dir);
}

###############################################################################
sub _process_file {
    my $self = shift;
    my ($filename, $directory) = @_;

	my $fullname = join '/', $directory, $filename;

    if (grep {$fullname =~ /$_/} @{$self->{include}}) {
        open(IN, $fullname) or croak "Cannot open '$fullname': $!\n";
        while ( my $line = <IN> ) {
            if ( my ($title) = ($line =~ /<h1.*?>(.*)<\/h1>/) ) {
                $self->{title}->{$fullname}  = $title;
                $self->{title}->{$directory} = $title if grep {$filename =~ /^$_$/} @{$self->{html_indices}};
                last;
            }
        }
        close(IN);
    }
    else {
    	$self->{title}->{$fullname} = $filename;
    }

    $self->{update}->{$fullname} = (stat($fullname))[9];
}

###############################################################################
sub _tie_cache {
    my $self = shift;

    tie(my %title,  'SDBM_File', "$self->{data_dir}/title",  O_RDWR|O_CREAT, 0666) or croak "Cache: Cannot tie SDBM file 'title': $!\n";
    tie(my %update, 'SDBM_File', "$self->{data_dir}/update", O_RDWR|O_CREAT, 0666) or croak "Cache: Cannot tie SDBM file 'update': $!\n";

    $self->{title}  = \%title;
    $self->{update} = \%update;
}

###############################################################################
sub _unlink_cache {
    my $self = shift;

    -e "$self->{data_dir}/title.dir" and do { unlink "$self->{data_dir}/title.dir" or croak "Cache: Cannot unlink 'title.dir': $!\n"; };
    -e "$self->{data_dir}/title.pag" and do { unlink "$self->{data_dir}/title.pag" or croak "Cache: Cannot unlink 'title.pag': $!\n"; };

    -e "$self->{data_dir}/update.dir" and do { unlink "$self->{data_dir}/update.dir" or croak "Cache: Cannot unlink 'update.dir': $!\n"; };
    -e "$self->{data_dir}/update.pag" and do { unlink "$self->{data_dir}/update.pag" or croak "Cache: Cannot unlink 'update.pag': $!\n"; };
}

###############################################################################
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) or croak "$self is not an object";

    my $name = $AUTOLOAD;
    $name =~ s/.*://;

    unless (exists $self->{_permitted}->{$name} ) {
        croak "Can't access '$name' field in class $type";
    }

    if (@_) {
        return $self->{$name} = shift;
    }
    else {
        return $self->{$name};
    }
}

###############################################################################
END {
    my $self = shift;

    untie(%{$self->{title}})  or carp "Cache: Cannot untie title cache\n";
    untie(%{$self->{update}}) or carp "Cache: Cannot untie update cache\n";
}

1;

###############################################################################
### DOCUMENTATION
###############################################################################

__END__

=head1 NAME

Yggdrasil::Cache - Manages titles and update times for a directory of HTML files

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 CONFIGURATION

=over 4

=item B<verbose=[n]>

Set verbosity level:

=over 4

=item -verbose=0

No output.

=item -verbose=1

Status messages.

=item -verbose=2

Print debug messages and file names.

=back

=back


=head1 BUGS

No known bugs.

=head1 AUTHOR

Morten Wulff <wulff@psyke.org>

Copyright (C) 2003 Morten Wulff

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=head1 VERSION

Version 1.1.2
