###############################################################################
### Bar.pm (Yggdrasil::Progress)
### Percent-bar progress indicator
###
### Copyright (C) 2003 Morten Wulff <wulff@psyke.org>
###
### This program is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License
### as published by the Free Software Foundation; either version 2
### of the License, or (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this program; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###############################################################################

### REVISIONS
### 1.0.0 : 2003-04-24 : Morten Wulff <wulff@psyke.org>
### 1.1.0 : 2003-08-25 : Morten Wulff <wulff@psyke.org>
###         Added support for longer markers (more than 1 char)

package Yggdrasil::Progress::Bar;

use strict;
use warnings;

$|++;

use Carp;
use Yggdrasil::Progress;

@Yggdrasil::Progress::Bar::ISA = qw(Yggdrasil::Progress);

###############################################################################
### INIT
###############################################################################

### default scale for the progress bar
my $Scale = <<SCALE;
0%  10%  20%  30%  40%  50%  60%  70%  80%  90%  100%
|----|----|----|----|----|----|----|----|----|----|
SCALE

### default settings
my %Fields = (
    ### used when the marker is in the 0-100 range
    marker => '>',

    ### used when the marker is out of range (i.e. $self->{counter) > 100
    clip => '<',

    ### 100 / width_of_scale (100/50=2 using the default scale)
    step => 2,

    ### the number of elements that constitute 100%
    elements => 0,

    ### use default scale
    scale => $Scale,

    ### optional char displayed before the marker to indicate mode/action/etc.
    char => ''
);

###############################################################################
### PUBLIC METHODS
###############################################################################

###############################################################################
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;

    my $self = bless($class->SUPER::new(), $class);

    foreach my $element (keys %Fields) {
        $self->{_permitted}->{$element} = $Fields{$element};
    }

    @{$self}{keys %Fields} = values %Fields;

    $self->{elements} = shift;

    ### check for presence of required parameter
    unless ( $self->{elements} && $self->{elements} > 0 ) {
        croak 'You need to specify a number of elements greater than zero';
    }

    ### initialize counters
    $self->{item}          = 0;
    $self->{scale_printed} = 0;
    $self->{clip_printed}  = 0;
    $self->{clip_counter}  = 1;

    return $self;
}

###############################################################################
sub update {
    my $self = shift;

    ### print scale if this is the first time we're called
    unless ($self->{scale_printed}) {
        print $self->{scale};
        print $self->{marker};
        $self->{scale_printed}++;
    }

    $self->{item}++;
    my $percentile = $self->{item} / $self->{elements} * 100;

    ### print marker
    while ($self->{counter} < $percentile) {
        $self->{counter} += $self->{step};
        if ($self->{counter} <= 100) {
            print join ' ', "\b" x length $self->{marker}, $self->{marker};
        }
    }

    ### print overflow marker (clip)
    if ($self->{counter} > 100) {
        unless ($self->{clip_printed}) {
            print join '',
                "\b" x length $self->{marker},
                ' '  x length $self->{marker},
                "$self->{clip}      ";
            $self->{clip_printed}++;
        }
        printf "\b\b\b\b\b\b%-6u", $self->{clip_counter};
        $self->{clip_counter}++;
    }
}

1;
